import 'dart:convert';

import 'package:bird/smart_api.dart';
import 'package:meta/meta.dart';

/// Classes for parameter/argument/Uri formatting
/// Currently supported
///     - Various formats for a POST Body
///     - JSONRPC 2.0 Request parser
///
/// These are helper functions for creating a request body.
///
/// Routes the creation of request bodies through a single object.
class BodyHelper {
  const BodyHelper();

  /// Sometimes an empty body is needed
  /// this explicitly says that
  String empty() => "";

  /// when a raw body is passed this explicitly says that
  String raw(String str) => str;

  /// Maps can be passed through here and will be converted to json strings
  String jsonObj(Map<String, dynamic> map) => json.encode(map);

  /// 2.0 JSONRPC methods can be created through here
  String rpc2_0({
    @required String method,
    @required int id,
    @required List<dynamic> params,
  }) =>
      JSONRPC2_0RequestObject(
        id: id,
        method: method,
        params: params,
      ).toString();
}
