import 'dart:async';
import 'dart:io';

import 'package:bird/smart_api.dart';
import 'package:bird/smart_json.dart';
import 'package:dartz/dartz.dart';
import 'package:test/test.dart';

class TestModel extends SmartJSON<TestModel> {
    final SmartString a = SmartString("aa");
    final SmartString b = SmartString("bb");

    @override
    TestModel creator() => TestModel();

    @override
    JSONPrototypeCodec<TestModel> tactic() =>
        JSONMapTactic.ignoreMissing({"a": a, "b": b});
}

class TestModel2 extends SmartJSON<TestModel2> {
    final SmartString x = SmartString("xx");
    final SmartString y = SmartString("yy");

    @override
    TestModel2 creator() => TestModel2();

    @override
    JSONPrototypeCodec<TestModel2> tactic() =>
        JSONMapTactic.ignoreMissing({"x": x, "y": y});
}

class TestModel3 extends SmartJSON<TestModel3> {
    final SmartModelList<TestModel> m = SmartModelList([
        TestModel()
            ..a.safeSet("x")
            ..b.safeSet("y"),
        TestModel()
            ..a.safeSet("xx")
            ..b.safeSet("yy"),
        TestModel()
            ..a.safeSet("xxx")
            ..b.safeSet("yyy"),
        TestModel()
            ..a.safeSet("xxxx")
            ..b.safeSet("yyyx"),
    ], () => TestModel());

    @override
    TestModel3 creator() => TestModel3();

    @override
    JSONPrototypeCodec<TestModel3> tactic() =>
        JSONMapTactic.ignoreMissing({"models": m});
}

class TestRestAPI extends APIRestDescription {

    const TestRestAPI() : super(host: "example.com", path: "api");

    APIRestMethodDescription get m => get();

    APIRestMethodDescription get exampleGetMinimal => get();

    APIRestMethodDescription get exampleGetFull =>
        get(
            contentType: [APIContentTypes.anyJson],
            path: "create",
            cmd: "CREATE",
            args: (a) => <String, dynamic>{"a": "b", "c": ["d", "e"]},
            headers: (method) => {"header-key": "header-value"},
        );

    APIRestMethodDescription get examplePostMinimal => post();

    APIRestMethodDescription get examplePostFull =>
        post(
            contentType: [APIContentTypes.anyJson],
            path: "create",
            cmd: "CREATE",
            body: (a, helper) => helper.raw("{}"),
            headers: (method) => {"header-key": "header-value"},
        );

    APIRestMethodDescription get exampleGetInvalidArguments =>
        get(
            contentType: [APIContentTypes.anyJson],
            path: "create",
            cmd: "CREATE",
            args: (a) => <String, dynamic>{"a": const Duration(minutes: 1), "c": ["d", "e"]},
            headers: (method) => {"header-key": "header-value"},
        );
}

class MockWebRequestService implements WebRequestService {

    Either<APIErrors, String> Function() onRequest;

    MockWebRequestService(this.onRequest);

    @override
    WebRequestInterface get requestInterface => MockWebRequestInterface(onRequest);
}

class MockWebRequestInterface implements WebRequestInterface {

    final Either<APIErrors, String> Function() onRequest;

    const MockWebRequestInterface(this.onRequest);

    @override
    Future<Either<APIErrors, String>> request(APIRestMethodDescription method) async {
        return onRequest();
    }
}

class CustomTestHeaders extends HttpHeaders {

    final Map<String, List<String>> _data;

    CustomTestHeaders(this._data);

    @override
    List<String> operator [](String name) => _data[name];

    @override void add(String name, Object value) {}

    @override void clear() {}

    @override void forEach(void Function(String name, List<String> values) f) {}

    @override void noFolding(String name) {}

    @override void remove(String name, Object value) {}

    @override void removeAll(String name) {}

    @override void set(String name, Object value) {}

    @override String value(String name) => null;
}

Matcher leftAPIErrorIs<T>() =>
    const TypeMatcher<Left<APIErrors, String>>()
        .having((a) => a.fold(id, (a) => null), "should be $T", TypeMatcher<T>());

Matcher leftDataErrorIs<T>() =>
    const TypeMatcher<Left<StorageErrors, String>>()
        .having((a) => a.fold(id, (a) => null), "should be $T", TypeMatcher<T>());

Matcher leftAllErrorIsWith<T, R>() =>
    TypeMatcher<Left<AllErrors, R>>()
        .having((a) => a.fold(id, (a) => null), "should be $T", TypeMatcher<T>());

Matcher leftAPIErrorIsWithRight<T, R>() =>
    TypeMatcher<Left<APIErrors, R>>()
        .having((a) => a.fold(id, (a) => null), "should be $T", TypeMatcher<T>());

Matcher rightAPIErrorsDataIs<T>(T data) =>
    TypeMatcher<Right<APIErrors, T>>()
        .having((a) => a.fold((a) => null, (a) => a), "", data);

Matcher rightAllErrorsDataIs<T>(T data) =>
    TypeMatcher<Right<AllErrors, T>>()
        .having((a) => a.fold((a) => null, (a) => a), "", data);

Matcher rightDataErrorsDataIs<T>(T data) =>
    TypeMatcher<Right<StorageErrors, T>>()
        .having((a) => a.fold((a) => null, (a) => a), "", data);

Matcher throwsSymbolAssert(Symbol s) => throwsA(const TypeMatcher<AssertionError>().having((a) => a.message, "", s));