import 'dart:convert';
import 'package:bird/smart_json.dart';
import 'package:decimal/decimal.dart';
import 'package:test/test.dart';

void main() {
  group("converter", () {
    group("$IsJSONSmartCodec", () {
      test("from", () {
        final a = IsJSONSmartCodec<Map<String, String>>(fromJSON: (dynamic a) {
          final dynamic x = json.decode(json.encode(a));
          return Map<String, String>.from(x as Map);
        });
        expect(a.encode({"a": "b"}), {"a": "b"});
      });
      test("to", () {
        expect(
            IsJSONSmartCodec<SmartBool>(fromJSON: (dynamic a) => null).decode(
                SmartBool(true)) as bool, true);
        expect(IsJSONSmartCodec<_MapNull>(fromJSON: (dynamic a) => null).decode(
            _MapNull()), {"a": "123", "x": "123"});
        expect(
            IsJSONSmartCodec<Iterable<SmartBool>>(fromJSON: (dynamic a) => null)
                .decode([SmartBool(true), SmartBool(false)]), [true, false]);
        expect(IsJSONSmartCodec<Map<String, int>>(fromJSON: (dynamic a) => null)
            .decode({"a": 1, "b": 2}), {"a": 1, "b": 2});
      });

      test("toStringToJSON", () {
        expect(
          IsJSONSmartCodec<_TestEnum>(
            fromJSON: (dynamic a) =>
                firstOfList(_TestEnum.values, _TestEnum.a)(a?.toString()),
            useToStringToJson: true,
          ).decode(_TestEnum.a),
          "_TestEnum.a",
        );

        expect(
          IsJSONSmartCodec<_TestEnum>(
            fromJSON: (dynamic a) =>
                firstOfList(_TestEnum.values, _TestEnum.a)(a?.toString()),
            useToStringToJson: true,
          ).encode("_TestEnum.a"),
          _TestEnum.a,
        );

        /// Wrong, cant serialize that
        expect(
          IsJSONSmartCodec<_TestEnum>(
            fromJSON: (dynamic a) =>
                firstOfList(_TestEnum.values, _TestEnum.a)(a?.toString()),
          ).decode(_TestEnum.a),
          _TestEnum.a,
        );
      });
    });
    group("$PassThroughCodec", () {
      test("from", () {
        const a = PassThroughCodec<Map<String, String>>();
        expect(a.encode({"a": "b"}), {"a": "b"});
      });
      test("to", () {
        expect(const PassThroughCodec<SmartBool>().decode(SmartBool(true)),
            const TypeMatcher<SmartBool>());
        expect(const PassThroughCodec<_MapNull>().decode(_MapNull()),
            const TypeMatcher<_MapNull>());
        expect(const PassThroughCodec<List<SmartBool>>().decode(
            [SmartBool(true), SmartBool(true)]),
            const TypeMatcher<List<SmartBool>>());
        expect(
            const PassThroughCodec<Map<String, int>>().decode({"a": 1, "b": 2}),
            const TypeMatcher<Map<String, int>>());
      });
    });
    group("$ToStringCodec", () {
      test("from", () {
        final a = ToStringCodec<Decimal>(
            fromJSON: (dynamic a) => Decimal.tryParse(a as String));
        expect(a.encode("123"), Decimal.tryParse("123"));
      });
      test("to", () {
        expect(ToStringCodec<SmartBool>(fromJSON: (dynamic a) =>
            SmartBool(json.decode(a as String) as bool)).decode(
            SmartBool(true)), "true");
      });
    });
  });
}

class _MapNull extends SmartJSON<_MapNull> {
  final SmartString a = SmartString("123");
  final SmartString x = SmartString("123");

  @override
  _MapNull creator() => _MapNull();

  @override
  JSONPrototypeCodec<_MapNull> tactic() =>
      JSONMapTactic.nullOnMissing({"a": a, "x": x});
}

enum _TestEnum { a, b, c }