import 'package:bird/smart_json.dart';
import 'package:test/test.dart';

void main() {
  group("smartdictt", () {
    group("SmartDictT", () {
      final init = {
        "a": _TestEnum.two,
      };
      final t = SmartDictT<_TestEnum>(
        init,
        tCreator: (dynamic a) =>
            firstOfList(_TestEnum.values, _TestEnum.four)(a),
      );

      test("Initialisation", () {
        expect(t.nullVal(), init);
      });

      test("Mutation", () {
        t.nullVal()["b"] = _TestEnum.one;
        t.nullVal().remove("a");
        expect(t.nullVal(), {"b": _TestEnum.one});
      });

      test("jsonConverter", () {
        expect(t.codec().decode(init), init);
        expect(t.codec().encode(init), init);
      });

      group("no tCreator", () {
        final init = {
          "a": true,
        };
        final t = SmartDictT<bool>(init);

        test("jsonConverter", () {
          expect(t.codec().decode(init), init);
          expect(t.codec().encode(init), init);
        });
      });
    });

    group("SmartModelDict", () {
      final init = {
        "a": _List2(),
      };

      final t = SmartModelDict<_List2>(
        init, () => _List2(),
      );

      test("creator not null", () {
        expect(() => SmartModelDict<_List2>(init, null),
            throwsA(const TypeMatcher<AssertionError>()));
        expect(() => SmartModelDict<_List2>(init, () => null).tCreator(0),
            throwsA(const TypeMatcher<AssertionError>()));
      });

      test("Initialisation", () {
        expect(t.nullVal(), init);
      });

      test("toJson", () {
        expect(t.to(), {"a": [null, null, null, null]});
      });

      test("fromJson", () {
        expect(t.codec().encode({"a": [true, 3, 1.23, "str"]}), {"a": _List2()
          ..a.safeSet(true)
          ..b.safeSet(3)
          ..c.safeSet(1.23)
          ..d.safeSet("str")});
      });
    });

    group("SmartRegistry", () {
      final init = {
        "a": [_TestEnum.one, _TestEnum.two],
      };

      final t = SmartRegistry<_TestEnum>(
        init,
        tCreator: (dynamic a) =>
            firstOfList(_TestEnum.values, _TestEnum.four)(a),
      );

      test("Initialisation", () {
        expect(t.nullVal(), init);
      });

      test("Mutation", () {
        t.nullVal()["a"].remove(_TestEnum.one);
        expect(t.nullVal(), {"a": [_TestEnum.two]});
      });

      test("applyToggle", () {
        final t = SmartRegistry<_TestEnum>(
          {"a": [_TestEnum.one, _TestEnum.two]},
          tCreator: (dynamic a) =>
              firstOfList(_TestEnum.values, _TestEnum.four)(a),
        );
        t.applyToggle("a", _TestEnum.three);
        t.applyToggle("a", _TestEnum.one);
        expect(t.to(), {"a": [_TestEnum.two, _TestEnum.three]});
        expect(t.to(), {"a": [_TestEnum.two, _TestEnum.three]});
        t.applyToggle("b", _TestEnum.one);
        expect(t.to(),
            {"a": [_TestEnum.two, _TestEnum.three], "b": [_TestEnum.one]});
      });

      test("applyAdd", () {
        final t = SmartRegistry<_TestEnum>(
          {"a": [_TestEnum.one, _TestEnum.two]},
          tCreator: (dynamic a) =>
              firstOfList(_TestEnum.values, _TestEnum.four)(a),
        );
        t.applyAdd("a", _TestEnum.three);
        t.applyAdd("a", _TestEnum.three);
        t.applyAdd("a", _TestEnum.one);
        expect(t.to(), {"a": [_TestEnum.one, _TestEnum.two, _TestEnum.three]});
      });

      test("applyAddAll", () {
        final t = SmartRegistry<_TestEnum>(
          {"a": [_TestEnum.one, _TestEnum.two]},
          tCreator: (dynamic a) =>
              firstOfList(_TestEnum.values, _TestEnum.four)(a),
        );
        t.applyAddAll("a", [_TestEnum.one, _TestEnum.three, _TestEnum.three]);
        expect(t.to(), {"a": [_TestEnum.one, _TestEnum.two, _TestEnum.three]});
      });

      test("applyAddOrRemove", () {
        final t = SmartRegistry<_TestEnum>(
          {"a": [_TestEnum.one, _TestEnum.two]},
          tCreator: (dynamic a) =>
              firstOfList(_TestEnum.values, _TestEnum.four)(a),
        );
        t.applyAddOrRemove("a", _TestEnum.three, true);
        expect(t.to(), {"a": [_TestEnum.one, _TestEnum.two, _TestEnum.three]});
        t.applyAddOrRemove("a", _TestEnum.three, true);
        expect(t.to(), {"a": [_TestEnum.one, _TestEnum.two, _TestEnum.three]});
        t.applyAddOrRemove("a", _TestEnum.three, false);
        expect(t.to(), {"a": [_TestEnum.one, _TestEnum.two]});
        t.applyAddOrRemove("a", _TestEnum.three, false);
        expect(t.to(), {"a": [_TestEnum.one, _TestEnum.two]});
      });

      group("tCreator is null", () {
        final t = SmartRegistry<bool>({
          "a": [false, true, false],
        });
        test("Initialisation", () {
          expect(t.nullVal(), {"a": [false, true, false]});
        });

        test("toJson", () {
          expect(t.to(), {"a": [false, true, false]});
        });

        test("fromJson", () {
          expect(t.codec().encode({"a": [false, true, false]}),
              {"a": [false, true, false]});
        });
      });

      group("fromJson", () {
        final t = SmartRegistry<SmartBool>(
            {}, tCreator: (dynamic a) => SmartBool(a as bool)).codec().encode({
          "a": [false, true, false],
        });
        test("Initialisation", () {
          expect(
              t, {"a": [SmartBool(false), SmartBool(true), SmartBool(false)]});
        });
      });

      group("toJson", () {
        final t = SmartRegistry<SmartBool>({
          "a": [SmartBool(false), SmartBool(true), SmartBool(false)],
        });
        test("Initialisation", () {
          expect(t.to(), {"a": [false, true, false]});
        });
      });
    });

    group("SmartDomainIdValue", () {
      final init = {
        "a": {
          "aa": _TestEnum.one,
        }
      };

      final t = SmartDomainIdValue<_TestEnum>(
        init,
        tCreator: (dynamic a) =>
            firstOfList(_TestEnum.values, _TestEnum.four)(a),
      );

      test("Initialisation", () {
        expect(t.nullVal(), init);
      });

      test("Mutation", () {
        t.nullVal()["a"]["aa"] = _TestEnum.two;
        expect(t.nullVal(), {
          "a": {
            "aa": _TestEnum.two,
          }
        });
      });

      test("toJson", () {
        expect(t.to(), init);
      });

      test("fromJson", () {
        expect(t.codec().encode({"a": {"aa": _TestEnum.one}}),
            {"a": {"aa": _TestEnum.one}});
      });

      group("tCreator is null", () {
        final init = {"a": {"aa": true}};

        final t = SmartDomainIdValue<bool>(init);

        test("Innitialize", () {
          expect(t.nullVal(), init);
        });

        test("toJson", () {
          expect(t.to(), {"a": {"aa": true}});
        });

        test("fromJson", () {
          expect(t.codec().encode({"a": {"aa": true}}), {"a": {"aa": true}});
        });
      });
    });
  });
}

enum _TestEnum { one, two, three, four }

class _List2 extends SmartJSON<_List2> {
  final SmartBool a = SmartBool(null);
  final SmartInt b = SmartInt(null);
  final SmartDouble c = SmartDouble(null);
  final SmartString d = SmartString(null);

  @override
  _List2 creator() => _List2();

  @override
  JSONPrototypeCodec<_List2> tactic() => JSONListTactic([a, b, c, d]);
}