import 'dart:convert';

import 'package:bird/smart_json.dart';
import 'package:decimal/decimal.dart';
import 'package:test/test.dart';

void main() {
  group("smartlistt.dart", () {
    group("$SmartListT<String>", () {
      test("Initialized", () {
        expect(
            SmartListT<String>(["not ok", "ok2"]).nullVal(), ["not ok", "ok2"]);
      });

      test("applyToggle add", () {
        expect((SmartListT<String>(["not ok", "ok2"])
          ..applyToggle("good"))
            .nullVal(), ["not ok", "ok2", "good"]);
      });

      test("empty list", () {
        final a = SmartListT<dynamic>(<dynamic>[]);
        expect(a.nullVal(), <dynamic>[]);
      });

      test("Copy toggle remove", () {
        expect((SmartListT<String>(["not ok", "ok2"])
          ..applyToggle("ok2"))
            .nullVal(), ["not ok"]);
      });

      test("Copy add", () {
        expect((SmartListT<String>(["not ok", "ok2"])
          ..applyAdd("add"))
            .nullVal(), ["not ok", "ok2", "add"]);
      });

      test("Copy remove", () {
        expect((SmartListT<String>(["not ok", "ok2"])
          ..applyRemove("ok2"))
            .nullVal(), ["not ok"]);
      });

      test("applyRemoveWhere", () {
        expect((SmartListT<String>(["not ok", "ok2", "not not"])
          ..applyRemoveWhere((a) => a.startsWith("not")))
            .nullVal(), ["ok2"]);

        expect((SmartListT<String>(["not ok", "ok2", "not not"])
          ..applyRemoveWhere((a) => false))
            .nullVal(), ["not ok", "ok2", "not not"]);
      });

      test("toJson", () {
        expect(SmartListT<String>(["1", "2"]).to(), ["1", "2"]);
      });

      test("fromJson", () {
        expect(SmartListT<String>([]).codec().encode(["1", "2"]), ["1", "2"]);
      });
    });

    group("$SmartListT<_SmartEnumTestEnum>", () {
      final t = () =>
      SmartListT<_SmartEnumTestEnum>.converter(
        [_SmartEnumTestEnum.two, _SmartEnumTestEnum.three],
        listConverter: (dynamic value) {
          return List<dynamic>.of(value as List).map((dynamic f) =>
              firstOfList(_SmartEnumTestEnum.values, _SmartEnumTestEnum.one)(
                  f?.toString())).toList();
        },
        useToStringToJson: true,
      );

      test("Initialized", () {
        expect(
            t().nullVal(), [_SmartEnumTestEnum.two, _SmartEnumTestEnum.three]);
      });

      test("Copy toggle add", () {
        expect((t()
          ..applyToggle(_SmartEnumTestEnum.four))
            .nullVal(), [
          _SmartEnumTestEnum.two,
          _SmartEnumTestEnum.three,
          _SmartEnumTestEnum.four
        ]);

        expect((t()
          ..applyToggle(_SmartEnumTestEnum.two))
            .nullVal(), [_SmartEnumTestEnum.three]);
      });

      test("Copy toggle remove", () {
        expect((t()
          ..applyToggle(_SmartEnumTestEnum.three))
            .nullVal(), [_SmartEnumTestEnum.two]);
      });

      test("Copy add", () {
        expect((t()
          ..applyAdd(_SmartEnumTestEnum.four))
            .nullVal(), [
          _SmartEnumTestEnum.two,
          _SmartEnumTestEnum.three,
          _SmartEnumTestEnum.four
        ]);
      });

      test("Copy remove", () {
        expect((t()
          ..applyRemove(_SmartEnumTestEnum.two))
            .nullVal(), [_SmartEnumTestEnum.three]);
      });

      test("Copy remove at index", () {
        expect((t()
          ..applyRemoveAtIndex(0))
            .nullVal(), [_SmartEnumTestEnum.three]);
        expect((t()
          ..applyRemoveAtIndex(1))
            .nullVal(), [_SmartEnumTestEnum.two]);
        expect((t()
          ..applyRemoveAtIndex(2))
            .nullVal(), [_SmartEnumTestEnum.two, _SmartEnumTestEnum.three]);
      });

      test("Copy replace at index", () {
        expect((t()
          ..applyReplaceAtIndex(0, _SmartEnumTestEnum.one))
            .nullVal(), [_SmartEnumTestEnum.one, _SmartEnumTestEnum.three]);
        expect((t()
          ..applyReplaceAtIndex(1, _SmartEnumTestEnum.one))
            .nullVal(), [_SmartEnumTestEnum.two, _SmartEnumTestEnum.one]);
        expect((t()
          ..applyReplaceAtIndex(2, _SmartEnumTestEnum.one))
            .nullVal(), [_SmartEnumTestEnum.two, _SmartEnumTestEnum.three]);
      });

      test("toJson", () {
        expect(
            t().to(), ["_SmartEnumTestEnum.two", "_SmartEnumTestEnum.three"]);
      });

      test("fromJson", () {
        expect(t().codec().encode(
            ["_SmartEnumTestEnum.two", "_SmartEnumTestEnum.three"]),
            [_SmartEnumTestEnum.two, _SmartEnumTestEnum.three]);
      });

      test("toString", () {
        expect(t().toString(),
            '["_SmartEnumTestEnum.two","_SmartEnumTestEnum.three"]');
      });
    });

    group("$SmartListT<_SmartEnumTestEnum>.enums", () {
      final t = () =>
      SmartListT<_SmartEnumTestEnum>.enums(
        [_SmartEnumTestEnum.two, _SmartEnumTestEnum.three],
        _SmartEnumTestEnum.two,
        _SmartEnumTestEnum.values,
      );

      test("toString", () {
        expect(t().toString(),
            '["_SmartEnumTestEnum.two","_SmartEnumTestEnum.three"]');
      });

      test("toJson", () {
        expect(
            t().to(), ["_SmartEnumTestEnum.two", "_SmartEnumTestEnum.three"]);
      });

      test("fromJson", () {
        expect(t().codec().encode(
            ["_SmartEnumTestEnum.two", "_SmartEnumTestEnum.three"]),
            [_SmartEnumTestEnum.two, _SmartEnumTestEnum.three]);
      });
    });

    group("$SmartModelList", () {
      SmartModelList<TestModel> t() =>
          SmartModelList<TestModel>(
            [TestModel()],
                () => TestModel(),
          );

      test("toString", () {
        expect(t().toString(), '[{"a":"aa","b":"bb"}]');
      });

      test("toJson", () {
        expect(t().to(), [{"a": "aa", "b": "bb"}]);
      });

      test("fromJson", () {
        expect(t().codec().encode(
            [{"a": "aa", "b": "bb"}, {"a": "aa", "b": "bb"}]),
            [TestModel(), TestModel()]);
      });
    });

    group("$SmartModelListWithWithStartsWithEndsWithTrie", () {
      SmartModelListWithWithStartsWithEndsWithTrie<_ModelWithID> t;

      final one = _ModelWithID()
        ..id.safeSet("btceth")
        ..something.safeSet(1);
      final two = _ModelWithID()
        ..id.safeSet("btcnem")
        ..something.safeSet(2);
      final three = _ModelWithID()
        ..id.safeSet("xrbxrp")
        ..something.safeSet(3);
      final four = _ModelWithID()
        ..id.safeSet("ethltc")
        ..something.safeSet(4);
      final five = _ModelWithID()
        ..id.safeSet("xrpeth")
        ..something.safeSet(5);
      setUp(() {
        t = SmartModelListWithWithStartsWithEndsWithTrie(
          [], (a) => a.id.nullVal(), () => _ModelWithID(),
        );

        t.safeSet([one, two, three, four, five]);
      });

      group("getStartsWith", () {
        test("one letter", () {
          final a = t.getStartsWith("x");
          expect(a, [three, five]);
        });
        test("one result", () {
          final a = t.getStartsWith("eth");
          expect(a, [four]);
        });
        test("null", () {
          expect(() => t.getStartsWith(null),
              throwsA(const TypeMatcher<AssertionError>()));
        });
      });

      group("getEndsWith", () {
        test("one letter", () {
          final a = t.getEndsWith("h").toList();
          expect(a, [one, five]);
        });
        test("one result", () {
          final a = t.getEndsWith("ltc");
          expect(a, [four]);
        });
        test("null", () {
          expect(() => t.getEndsWith(null),
              throwsA(const TypeMatcher<AssertionError>()));
        });
      });
    });

    group("$SmartDecimalList", () {
      test("contains", () {
        final list = [Decimal.tryParse("123.123"), Decimal.tryParse("321.321")];
        expect(SmartDecimalList(list).nullVal(), list);
      });

      test("toJson", () {
        final list = [Decimal.tryParse("123.123"), Decimal.tryParse("321.321")];
        expect(SmartDecimalList(list).to(), list.map((a) => a.toString()));
      });

      test("tCreator null ", () {
        expect(
          SmartDecimalList([]).codec().encode([
            "bla",
            "123.123",
            "0.000000000000000000000000000000000000000000000001"
          ]),
          [
            null,
            Decimal.tryParse("123.123"),
            Decimal.tryParse(
                "0.000000000000000000000000000000000000000000000001")
          ],
        );
        expect(
          SmartDecimalList([]).codec().decode(
              [
                null,
                Decimal.tryParse("123.123"),
                Decimal.tryParse(
                    "0.000000000000000000000000000000000000000000000001")
              ]),
          [
            null,
            "123.123",
            "0.000000000000000000000000000000000000000000000001"
          ],
        );
      });

      test("toString", () {
        final list = [Decimal.tryParse("123.123"), Decimal.tryParse("321.321")];
        expect(SmartDecimalList(list).toString(),
            json.encode(list.map((a) => a.toString()).toList()));
      });

      test("invalid becomes null toJson", () {
        expect(SmartDecimalList(
            [Decimal.tryParse("123.123"), Decimal.tryParse("b")]).to(),
            ["123.123", null]);
      });
    });
  });
}

enum _SmartEnumTestEnum { one, two, three, four }

class TestModel extends SmartJSON<TestModel> {
  final SmartString a = SmartString("aa");
  final SmartString b = SmartString("bb");

  @override
  TestModel creator() => TestModel();

  @override
  JSONPrototypeCodec<TestModel> tactic() =>
      JSONMapTactic.ignoreMissing({"a": a, "b": b});
}

class _ModelWithID extends SmartJSON<_ModelWithID> {

  final SmartString id = SmartString(null);
  final SmartInt something = SmartInt(null);

  @override
  _ModelWithID creator() => _ModelWithID();

  @override
  JSONMapTactic<_ModelWithID> tactic() =>
      JSONMapTactic.ignoreMissing({
        "id": id,
        "something": something,
      });
}