import 'package:bird/smart_json.dart';
import 'package:test/test.dart';

void main() {
  group("$SmartRedirect", () {
    final SmartRedirect redirect = SmartRedirect<dynamic>((dynamic a) {
      if (a["type"] == "string") {
        return SmartString("abc");
      }
      if (a["type"] == "int") {
        return SmartInt(123);
      }
      if (a["type"] == "double") {
        return SmartDouble(0.123);
      }
      return SmartBool(false);
    });
    test("string", () {
      expect(redirect
          .from({"type": "string"})
          .access, SmartString("abc"));
    });
    test("double", () {
      expect(redirect
          .from({"type": "double"})
          .access, SmartDouble(0.123));
    });
    test("int", () {
      expect(redirect
          .from({"type": "int"})
          .access, SmartInt(123));
    });
    test("bla", () {
      expect(redirect
          .from({"type": "bla"})
          .access, SmartBool(false));
    });

    final r = SmartRedirect(TestModel().from).from(TestModel().to());

    test("parsing", () {
      expect(r
          .from(r.to())
          .access, TestModel());
    });
  });

  group("$SmartInherit", () {
    test("deserialize nonsense", () {
      final a = {"type": "Somethingsomething", "hello": ["aabc", "abd"]};
      expect(
        _InheritedParent()
            .from({"inherited": a})
            .inherited
            .nullVal()
            .access
            .to(),
        {"type": "_InheritedA", "hello": "world"},
      );
    });
    test("a", () {
      final a = {"type": "_InheritedA", "hello": "me"};
      expect(
        _InheritedParent()
            .from({"inherited": a})
            .inherited
            .nullVal()
            .access
            .to(),
        a,
      );
    });
    test("b", () {
      final a = {"type": "_InheritedB", "hello": "them", "on": true};
      expect(
        _InheritedParent()
            .from({"inherited": a})
            .inherited
            .nullVal()
            .access
            .to(),
        a,
      );
    });
    test("c with custom type identifier", () {
      final a = {"type": "c", "howmany": 123};
      expect(
        _InheritedParent()
            .from({"inherited": a})
            .inherited
            .nullVal()
            .access
            .to(),
        a,
      );
    });
  });
}

class TestModel extends SmartJSON<TestModel> {
  final SmartString a = SmartString("aa");
  final SmartString b = SmartString("bb");

  @override
  TestModel creator() => TestModel();

  @override
  JSONPrototypeCodec<TestModel> tactic() =>
      JSONMapTactic.ignoreMissing({"a": a, "b": b});
}

class _InheritedParent extends SmartJSON<_InheritedParent> {

  final SmartInherit<_Inherited> inherited = SmartInherit<_Inherited>(
      _InheritedA(), [
    _InheritedA(),
    _InheritedB(),
    _InheritedC(),
  ]);

  @override
  _InheritedParent creator() => _InheritedParent();

  @override
  JSONMapTactic<_InheritedParent> tactic() =>
      JSONMapTactic.ignoreMissing({
        "inherited": inherited,
      });
}

abstract class _Inherited extends InheritableSmartJSON<_Inherited> {}

class _InheritedA extends _Inherited {
  final SmartString hello = SmartString("world");

  @override
  _Inherited creator() => _InheritedA();

  @override
  Map<String, JSONSmartVariableInterface> mapTacticVars() => {"hello": hello};
}

class _InheritedB extends _Inherited {
  final SmartString hello = SmartString("you");
  final SmartBool on = SmartBool(false);

  @override
  _InheritedB creator() => _InheritedB();

  @override
  Map<String, JSONSmartVariableInterface> mapTacticVars() =>
      {"hello": hello, "on": on};
}

class _InheritedC extends _Inherited {
  final SmartInt howmany = SmartInt(0);

  @override
  _InheritedC creator() => _InheritedC();

  @override
  String typeSerializationString() => "c";

  @override
  Map<String, JSONSmartVariableInterface> mapTacticVars() =>
      {"howmany": howmany};
}