import 'package:bird/smart_json.dart';
import 'package:dartz/dartz.dart' hide State;
import 'package:decimal/decimal.dart';
import 'package:test/test.dart';

void main() {
  group("smartt.dart", () {
    group("$SmartTFrom", () {
      test("not null", () {
        expect(() => SmartTFrom(null),
            throwsA(const TypeMatcher<AssertionError>()));
      });

      test("toJson", () {
        expect(SmartTFrom(_List2()).to(), [null, null, null, null]);
      });

      test("fromJson", () {
        expect(SmartTFrom(_List2()).codec().encode([null, null, null, null]),
            _List2());
      });
    });

    group("$SmartBool", () {
      final t = SmartBool(false);

      test("initialized", () {
        expect(t.nullVal(), false);
      });

      test("apply", () {
        final t = SmartBool(false);
        expect(t.apply((a) => !a), true);
      });

      test("apply", () {
        final t = SmartBool(false);
        expect(t.safeApply(() => null, (a) => !a), true);

        final t2 = SmartBool(null);
        expect(t2.safeApply(() => null, (a) => !a), null);
      });

      test("applyToggle on null", () {
        t.safeSet(null);
        expect((t..applyToggle())
            .nullVal(), true);
      });

      test("applyToggle on value", () {
        t.safeSet(false);
        expect((t..applyToggle()).nullVal(), true);
      });

      test("==", () {
        expect(SmartBool(true), SmartBool(true));
        expect(SmartBool(true), isNot(SmartBool(false)));
      });

      test("hashCode", () {
        expect(SmartBool(true).hashCode, SmartBool(true).hashCode);
        expect(SmartBool(true).hashCode, isNot(SmartBool(false).hashCode));
      });
    });

    group("$JSONSmartVariableInterface option", () {
      test("none", () {
        final t = SmartBool(null);
        expect(t.option, const TypeMatcher<None>());
      });

      test("some", () {
        final t = SmartBool(true);
        expect(t.option, const TypeMatcher<Some>());
      });
    });

    group("$SmartInt", () {
      final t = SmartInt(1);

      test("initialized", () {
        expect(t.nullVal(), 1);
      });

      test("increment on null", () {
        t.safeSet(null);
        expect((t..applyIncrement()).nullVal(), 1);
      });

      test("increment on value", () {
        t.safeSet(4);
        expect((t..applyIncrement()).nullVal(), 5);
      });

      test("applyNowIfNot", () {
        final t = SmartInt(null);
        expect((t..applyNowIfNot()).nullVal(), isNotNull);
        final a = t.nullVal();
        expect((t..applyNowIfNot()).nullVal(), a);
        final t2 = SmartInt(-1);
        expect((t2..applyNowIfNot()).nullVal(), isNotNull);
      });

      test("copyNow", () {
        expect(DateTime
            .fromMillisecondsSinceEpoch((t..applyNow()).nullVal())
            .difference(DateTime.now())
            .inMilliseconds, lessThan(1000));
      });
    });

    group("$SmartDouble", () {
      final t = SmartDouble(3.3);

      test("initialized", () {
        expect(t.nullVal(), 3.3);
      });
    });

    group("$SmartString", () {
      final t = SmartString("nothi");

      test("initialized", () {
        expect(t.nullVal(), "nothi");
      });
    });

    group("$SmartDynamic", () {
      final t = SmartDynamic("nothi");

      test("initialized", () {
        expect(t.nullVal(), "nothi");
      });
    });

    group("$SmartTRawFrom", () {
      final a = SmartTRawFrom(Decimal.tryParse("1.0"), (dynamic a) =>
          Decimal.tryParse(a as String));
      test("jsonConverter", () {
        expect(a.codec().encode("1.0"), Decimal.tryParse("1.0"));
        expect(a.codec().decode(Decimal.tryParse("1.0")).toString(), "1");
      });
    });

    group("$SmartLazyDecimal lazy", () {
      final t = SmartLazyDecimal("1.0");

      test("isLazy by default", () {
        expect(t.isLazy, true);
      });

      test("initialized", () {
        expect(t.decimal, Decimal.tryParse("1.0"));
      });

      test("safeSet invalid", () {
        expect((t..safeSet("t")).decimal, null);
      });

      test("safeSet valid", () {
        expect((t..safeSet("1")).decimal, Decimal.tryParse("1.0"));
      });

      test("toJson", () {
        expect(t.to(), "1");
      });

      test("fromJson", () {
        expect(t.codec().encode("1.0"), "1.0");
      });
    });

//        group("$SmartLazyBigInt lazy", () {
//            String source;
//
//            SmartLazyBigInt t;
//
//            setUp(() {
//                source = List.filled(40, (a) => "9").join();
//                t = SmartLazyBigInt(source);
//            });
//
//            test("isLazy by default", () {
//                expect(t.isLazy, true);
//            });
//
//            test("initialized", () {
//                expect(t.bigInt, BigInt.tryParse(source));
//            });
//
//            test("safeSet invalid", () {
//                expect((t..safeSet("t")).bigInt, null);
//            });
//
//            test("safeSet valid", () {
//                expect((t..safeSet("1")).bigInt, BigInt.tryParse("1"));
//            });
//
//            test("toJson", () {
//                expect(t.toJson(), source);
//            });
//
//            test("fromJson", () {
//                expect(t.jsonConverter().from("1"), "1");
//            });
//        });

    group("$SmartLazyDecimal not lazy", () {
      final t = SmartLazyDecimal("1.0", isLazy: false);

      test("not lazy", () {
        expect(t.isLazy, false);
      });

      test("initialized", () {
        expect(t.decimal, Decimal.tryParse("1.0"));
      });

      test("safeSet invalid", () {
        expect((t..safeSet("t")).decimal, null);
      });

      test("safeSet valid", () {
        expect((t..safeSet("1")).decimal, Decimal.tryParse("1.0"));
      });

      test("toJson", () {
        expect(t.to(), "1");
      });

      test("fromJson", () {
        expect(t.codec().encode("1.0"), "1.0");
      });
    });

    group("$SmartNonNullableEnum", () {
      final t = SmartNonNullableEnum<_SmartEnumTestEnum>(
          _SmartEnumTestEnum.two, _SmartEnumTestEnum.values);

      test("initialized", () {
        expect(t.isThis(_SmartEnumTestEnum.two), true);
        t.safeSet(null);
        expect(t.nullVal(), _SmartEnumTestEnum.two);
      });

      test("can't be initialized to null", () {
        expect(() => SmartNonNullableEnum<_SmartEnumTestEnum>(
            null, _SmartEnumTestEnum.values),
            throwsA(const TypeMatcher<AssertionError>()));
        expect(() => SmartNonNullableEnum<_SmartEnumTestEnum>(
            _SmartEnumTestEnum.two, null),
            throwsA(const TypeMatcher<AssertionError>()));
      });

      test("toJson", () {
        expect(SmartNonNullableEnum<_SmartEnumTestEnum>(
            _SmartEnumTestEnum.one, _SmartEnumTestEnum.values).to(),
            "_SmartEnumTestEnum.one");
      });

      test("fromJson", () {
        expect(
            SmartNonNullableEnum<_SmartEnumTestEnum>(
                _SmartEnumTestEnum.one, _SmartEnumTestEnum.values)
                .codec()
                .encode(
                "_SmartEnumTestEnum.one"),
            _SmartEnumTestEnum.one);
      });

      test("fromJson invalid becomes init", () {
        expect(SmartNonNullableEnum<_SmartEnumTestEnum>(
            _SmartEnumTestEnum.one, _SmartEnumTestEnum.values).codec().encode(
            "_SmartEnumTestEnum.removed"), _SmartEnumTestEnum.one);
        expect(SmartNonNullableEnum<_SmartEnumTestEnum>(
            _SmartEnumTestEnum.one, _SmartEnumTestEnum.values).codec().encode(
            "123"),
            _SmartEnumTestEnum.one);
        expect(SmartNonNullableEnum<_SmartEnumTestEnum>(
            _SmartEnumTestEnum.one, _SmartEnumTestEnum.values).codec().encode(
            null),
            _SmartEnumTestEnum.one);
      });
    });
  });
}

enum _SmartEnumTestEnum { one, two, three }

class _List2 extends SmartJSON<_List2> {
  final SmartBool a = SmartBool(null);
  final SmartInt b = SmartInt(null);
  final SmartDouble c = SmartDouble(null);
  final SmartString d = SmartString(null);

  @override
  _List2 creator() => _List2();

  @override
  JSONPrototypeCodec<_List2> tactic() => JSONListTactic([a, b, c, d]);
}